;; Generated package description from extmap.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "extmap" "1.3.10.20250203.193959" "Externally-stored constant mapping for Elisp" '((emacs "24.4")) :commit "bb4867e13ec7480d4dbc1d99888627e4cc8ef2f6" :authors '(("Paul Pogonyshev" . "pogonyshev@gmail.com")) :maintainer '("Paul Pogonyshev" . "pogonyshev@gmail.com") :keywords '("lisp") :url "https://github.com/doublep/extmap")
